/* UNH CS 758 assignment 7

*/

#include <stdio.h>
#include <stdlib.h>  /* free, EXIT_SUCCESS, rand */
#include <string.h>  /* strtok */
#include <ctype.h>   /* islower, isspace */
#include "utils.h"
#include "parse.h"


/* for grammar file */
#define MAX_LINE_LEN 256
#define WHITESPACE " \a\b\f\n\r\t\v"

/* for sentences */
#define MAX_S_LEN 256


/************************************************************
 * making and freeing grammars
 ************************************************************/


cfg *empty_grammar(void)
{
  cfg *g = (cfg *) safe_malloc(sizeof(cfg), "can't malloc grammar");
  g->terminal_rules = NULL;
  g->nonterminal_rules = NULL;
  return g;
}


void add_term_rule(cfg *g, char *nonterm, char *term, double prob)
{
  terminal_rule *r = (terminal_rule *) safe_malloc(sizeof(terminal_rule),
						   "can't malloc new terminal rule!");
  r->nonterminal = strcpy_new(nonterm);
  r->terminal = strcpy_new(term);
  r->probability = prob;
  r->next = g->terminal_rules;
  g->terminal_rules = r;
}


void add_nonterm_rule(cfg *g, char *nonterm, char *nt1, char *nt2, double prob)
{
  nonterminal_rule *r = (nonterminal_rule *) safe_malloc(sizeof(nonterminal_rule),
						       "can't malloc new nonterminal rule!");
  r->nonterminal = strcpy_new(nonterm);
  r->nonterminal_1 = strcpy_new(nt1);
  r->nonterminal_2 = strcpy_new(nt2);
  r->probability = prob;
  r->next = g->nonterminal_rules;
  g->nonterminal_rules = r;
}


void free_t_rules(terminal_rule *r)
{
  if (r) {
    terminal_rule *next = r->next;

    free(r->nonterminal);
    free(r->terminal);
    free(r);
    free_t_rules(next);
  }
}


void free_nt_rules(nonterminal_rule *r)
{
  if (r) {
    nonterminal_rule *next = r->next;

    free(r->nonterminal);
    free(r->nonterminal_1);
    free(r->nonterminal_2);
    free(r);
    free_nt_rules(next);
  }
}


void free_grammar(cfg *g)
{
  free_t_rules(g->terminal_rules);
  free_nt_rules(g->nonterminal_rules);
  free(g);
}


/************************************************************
 * read grammar
 ************************************************************/


/* skips over whitespace characters in infile.  returns EOF iff EOF */
int skip_white(FILE *infile)
{
  int c;
  
  do {
    c = fgetc(infile);
  } while (isspace(c));
  if (c != EOF) {
    ungetc(c, infile);
  }
  return c;
}


/* read through next \n or EOF, return EOF iff EOF */
int skip_line(FILE *infile)
{
  int c;
  
  do {
    c = fgetc(infile);
  } while ((c != EOF) && (c != '\n'));
  return c;
}


/* returns EOF iff EOF or error */
int skip_noncontent_lines(FILE *infile)
{  
  if (skip_white(infile) == EOF)
    return EOF;
  else {
    int c = fgetc(infile);
    
    if (c == '#') {
      if (skip_line(infile) == EOF)
	return EOF;
      else
	return skip_noncontent_lines(infile);
    }
    return ungetc(c, infile);
  }
}


/* returns pointer to static buffer holding next line of input, or
   NULL iff EOF or error.  lines starting with # or whitespace are skipped */
char *read_line(FILE *infile)
{
  static char buf[MAX_LINE_LEN+1];
  
  if (skip_noncontent_lines(infile) == EOF)
    return NULL;
  return fgets(buf, MAX_LINE_LEN, infile);
}


double get_prob()
{
  char *tok = strtok(NULL, WHITESPACE);

  if (! tok) panic("didn't find probability at end of grammar rule");
  return atof(tok);
}


/* returns 0 on EOF */
int read_rule(FILE *infile, cfg *g)
{
  char *l = read_line(infile);

  if (l) {
    char *nonterm = strtok(l, WHITESPACE);
    char *tok;

    if (! nonterm) panic("didn't find first token in grammar rule");
    tok = strtok(NULL, WHITESPACE);
    if (! tok) panic("didn't find second token in grammar rule");
    if (islower(*tok)) {
      /* terminal rule */
      add_term_rule(g, nonterm, tok, get_prob());
    } else {
      /* nonterminal rule */
      char *nt2 = strtok(NULL, WHITESPACE);
      if (! nt2) panic("didn't find second RHS nonterminal in grammar rule");
      add_nonterm_rule(g, nonterm, tok, nt2, get_prob());
    }
    return 1;
  } else
    return 0;
}


cfg *read_grammar(FILE *infile)
{
  cfg *g = empty_grammar();

  while (read_rule(infile, g)) {
  }
  return g;
}


/************************************************************
 * print grammar
 ************************************************************/


void fprint_terminal_rules(FILE *outfile, terminal_rule *r)
{
  if (r) {
    fprintf(outfile, "%s -> %s  %6f\n",
	    r->nonterminal, r->terminal, r->probability);
    fprint_terminal_rules(outfile, r->next);
  }
}


void fprint_nonterminal_rules(FILE *outfile, nonterminal_rule *r)
{
  if (r) {
    fprintf(outfile, "%s -> %s %s  %6f\n",
	    r->nonterminal, r->nonterminal_1, r->nonterminal_2,
	    r->probability);
    fprint_nonterminal_rules(outfile, r->next);
  }
}


void fprint_grammar(FILE *outfile, cfg *g)
{
  fprintf(outfile, "-- CFG --\n");
  fprint_terminal_rules(outfile, g->terminal_rules);
  fprint_nonterminal_rules(outfile, g->nonterminal_rules);
  fprintf(outfile, "-- end CFG --\n");
}


/************************************************************
 * the sentence
 ************************************************************/


sentence *empty_sentence()
{
  sentence *s = (sentence *) safe_malloc(sizeof(sentence),"can't malloc sentence");
  char **toks = (char **) safe_malloc(sizeof(char *) * MAX_S_LEN, "can't malloc toks array");
  s->len = 0;
  s->tokens = toks;
  return s;
}


sentence *read_sentence(FILE *infile, cfg *g)
{
  sentence *s = empty_sentence();
  char buf[MAX_S_LEN];
  char *tok;
  
  if (! fgets(buf, MAX_S_LEN, infile)) panic("EOF or error reading sentence");
  tok = strtok(buf, WHITESPACE);
  while (tok != NULL) {
    s->tokens[s->len++] = strcpy_new(tok);
    tok = strtok(NULL, WHITESPACE);
  }
  return s;
}


void fprint_sentence(FILE *outfile, sentence *s)
{
  int i;

  fprintf(outfile, "sentence (%d):", s->len);
  for (i = 0; i < s->len; i++) {
    fprintf(outfile, " %s", s->tokens[i]);
  }
  fprintf(outfile, "\n");
}


void free_sentence(sentence *s)
{
  while (s->len--) {
    free(s->tokens[s->len]);
  }
  free(s->tokens);
  free(s);
}


/************************************************************
 * parses
 ************************************************************/


/* recursive helper for fprint_parse.  p cover s from indices start to
   end inclusive */
void fp_parse(FILE *outfile, parse *p, sentence *s, int start, int end)
{
  if (p) {
    fprintf(outfile, "%s(", p->nonterminal);
    if (p->nonterminal_1) {
      if (start == end) panic("nt_1 <> NULL but start == end!");
      fp_parse(outfile, p->nonterminal_1, s, start, p->breakpoint);
      fprintf(outfile, " ");
      fp_parse(outfile, p->nonterminal_2, s, p->breakpoint+1, end);
    } else {
      if (start != end) panic("nt_1 = NULL but start <> end!");
      if (start != p->breakpoint) panic("nt_1 = NULL and start==end but start<>breakpoint");
      fprintf(outfile, "%s", s->tokens[start]);
    }
    fprintf(outfile, ")");
  }
}


void fprint_parse(FILE *outfile, parse *p, sentence *s, int start, int end)
{
  if (p == NULL) panic("tried to print NULL parse");
  fp_parse(outfile, p, s, start, end);
  fprintf(outfile, " %6f\n", p->probability);
}


/* prints parses of s from start to end inclusive */
void fprint_partial_parses(FILE *outfile, parse *ps, sentence *s, int start, int end)
{
  if (ps == NULL)
    fprintf(outfile, "No parse found!\n");
  else {
    while (ps != NULL) {
      fprint_parse(outfile, ps, s, start, end);
      ps = ps->next;
    }
  }
}

void fprint_parses(FILE *outfile, parse *ps, sentence *s)
{
  fprint_partial_parses(outfile, ps, s, 0, s->len - 1);
}


void free_parses_shallow(parse *ps)
{
  while (ps != NULL) {
    parse *next = ps->next;
    free(ps);
    ps = next;
  }
}


void free_parses_deep(parse *ps)
{
  while (ps != NULL) {
    parse *next = ps->next;
    free_parses_deep(ps->nonterminal_1);
    free_parses_deep(ps->nonterminal_2);
    free(ps);
    ps = next;
  }
}


/* EOF */
